@@ -1,13 +1,23 @@
 /*
  * listing.h
  *
- * Copyright 2000,2008 Werner Fink, 2000 SuSE GmbH Nuernberg, Germany.
- *				    2008 SuSE Linux Products GmbH Nuernberg, Germany
+ * Copyright 2000,2009 Werner Fink, 2000 SuSE GmbH Nuernberg, Germany.
+ *				    2008,2009 SuSE Linux Products GmbH Nuernberg, Germany
  *
  * This source is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
+ *
  */
 
 #include <stddef.h>
@@ -92,7 +102,7 @@ static inline void prefetch(const void *restrict x)
 #endif
 #define __packed attribute((packed))
 
-#define alignof(type)		(sizeof(type)+(sizeof(type)%sizeof(void*)))
+#define alignof(type)		((sizeof(type)+(sizeof(void*)-1)) & ~(sizeof(void*)-1))
 #define strsize(string)		((strlen(string)+1)*sizeof(char))
 
 typedef struct list_struct {
@@ -283,7 +293,7 @@ extern void lsort(const char type);
 extern const char *const delimeter;
 extern void error(const char *restrict fmt, ...) attribute((noreturn,format(printf,1,2)));
 extern void warn (const char *restrict fmt, ...) attribute((format(printf,1,2)));
-extern void info (const char *restrict fmt, ...) attribute((format(printf,1,2)));
+extern void info (int level, const char *restrict fmt, ...) attribute((format(printf,2,3)));
 extern inline int map_has_runlevels(void) attribute((always_inline));
 extern inline char map_runlevel_to_key(const int runlevel);
 extern inline ushort map_key_to_lvl(const char key);
@@ -307,42 +317,42 @@ static inline char * xstrdup(const char *restrict s)
 #define xreset(ptr)	\
 	{char *restrict tmp = (char *restrict)ptr; if (ptr && *tmp) free(ptr);} ptr = NULL
 
-#if defined(HAS_unlinkat) && defined(_ATFILE_SOURCE)
+#if defined(HAS_unlinkat) && defined(_ATFILE_SOURCE) && !defined(__stub_unlinkat)
 # define xremove(d,x) (__extension__ ({ if ((dryrun ? 0 : \
 	(unlinkat(d,x,0) != 0 && (errno != EISDIR || unlinkat(d,x,AT_REMOVEDIR) != 0)))) \
 	warn ("can not remove(%s%s): %s\n", rcd, x, strerror(errno)); \
 	else \
-	info("remove service %s/%s%s\n", path, rcd, x); }))
+	info(1, "remove service %s/%s%s\n", path, rcd, x); }))
 #else
 # define xremove(d,x) (__extension__ ({ if ((dryrun ? 0 : (remove(x) != 0))) \
 	warn ("can not remove(%s%s): %s\n", rcd, x, strerror(errno)); \
 	else \
-	info("remove service %s/%s%s\n", path, rcd, x); }))
+	info(1, "remove service %s/%s%s\n", path, rcd, x); }))
 #endif
-#if defined(HAS_symlinkat) && defined(_ATFILE_SOURCE)
+#if defined(HAS_symlinkat) && defined(_ATFILE_SOURCE) && !defined(__stub_symlinkat)
 # define xsymlink(d,x,y) (__extension__ ({ if ((dryrun ? 0 : (symlinkat(x, d, y) != 0))) \
 	warn ("can not symlink(%s, %s%s): %s\n", x, rcd, y, strerror(errno)); \
 	else \
-	info("enable service %s -> %s/%s%s\n", x, path, rcd, y); }))
+	info(1, "enable service %s -> %s/%s%s\n", x, path, rcd, y); }))
 #else
 # define xsymlink(d,x,y) (__extension__ ({ if ((dryrun ? 0 : (symlink(x, y) != 0))) \
 	warn ("can not symlink(%s, %s%s): %s\n", x, rcd, y, strerror(errno)); \
 	else \
-	info("enable service %s -> %s/%s%s\n", x, path, rcd, y); }))
+	info(1, "enable service %s -> %s/%s%s\n", x, path, rcd, y); }))
 #endif
-#if defined(HAS_fstatat) && defined(_ATFILE_SOURCE)
+#if defined(HAS_fstatat) && defined(_ATFILE_SOURCE) && !defined(__stub_fstatat)
 # define xstat(d,x,s)	(__extension__ ({ fstatat(d,x,s, 0); }))
 # define xlstat(d,x,s)	(__extension__ ({ fstatat(d,x,s, AT_SYMLINK_NOFOLLOW); }))
 #else
 # define xstat(d,x,s)	(__extension__ ({ stat(x,s); }))
 # define xlstat(d,x,s)	(__extension__ ({ lstat(x,s); }))
 #endif
-#if defined(HAS_readlinkat) && defined(_ATFILE_SOURCE)
+#if defined(HAS_readlinkat) && defined(_ATFILE_SOURCE) && !defined(__stub_readlinkat)
 # define xreadlink(d,x,b,l)	(__extension__ ({ readlinkat(d,x,b,l); }))
 #else
 # define xreadlink(d,x,b,l)	(__extension__ ({ readlink(x,b,l); }))
 #endif
-#if defined(HAS_openat) && defined(_ATFILE_SOURCE)
+#if defined(HAS_openat) && defined(_ATFILE_SOURCE) && !defined(__stub_openat)
 # define xopen(d,x,f)	(__extension__ ({ openat(d,x,f); }))
 #else
 # define xopen(d,x,f)	(__extension__ ({ open(x,f); }))
@@ -368,6 +378,7 @@ static inline char * xstrdup(const char *restrict s)
 #define SERV_SCRIPT	0x0080
 #define SERV_NOSTOP	0x0100
 #define SERV_CMDLINE	0x0200
+#define SERV_FIRST	0x0400
 
 /*
  * Bits of the runlevels
